#
# This file is part of Python package: `h5yaml`
#
#     https://github.com/rmvanhees/h5_yaml.git
#
# Copyright (c) 2025-2026 - R.M. van Hees (SRON)
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Define version of the software."""

__all__ = ["__version__"]


from ._version import __version__



#
# This file is part of Python package: `h5yaml`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025-2026 - R.M. van Hees (SRON)
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Initialise netCDF4 file from a YAML configuration file using `h5py` or `netCDF4`."""

from __future__ import annotations

__all__ = ["NcFromYaml"]

# import logging
import pprint
from pathlib import Path

import yaml

from .h5_create import H5Create
from .nc_create import NcCreate


# - local function -------------------------------------
def _from_yaml(file_path: Path | str) -> dict:
    """Read settings from a YAML file: `file_path`.

    Parameters
    ----------
    file_path :  Path | str
       full path of YAML file

    Returns
    -------
    dict
       content of the configuration file

    """
    if isinstance(file_path, str):
        file_path = Path(file_path)

    if not file_path.is_file():
        raise FileNotFoundError(f"{file_path} not found")

    with file_path.open("r", encoding="ascii") as fid:
        try:
            settings = yaml.safe_load(fid)
        except yaml.parser.ParserError as exc:
            raise RuntimeError(f"Failed to parse {file_path}") from exc

    return settings


# - class definition -----------------------------------
class NcFromYaml(H5Create):
    """Class to create a netCDF4 formated file from a YAML configuration file.

    Parameters
    ----------
    nc_yaml_fl :  Path | str | list[Path | str]
       YAML files with the HDF5 format definition

    """

    def __init__(
        self: NcFromYaml,
        nc_yaml_fl: Path | str | list[Path | str],
    ) -> None:
        """Construct a NcFromYaml instance."""
        # self.logger = logging.getLogger("h5yaml.NcFromYaml")
        self.module = "h5py"
        super().__init__()

        for yaml_fl in nc_yaml_fl if isinstance(nc_yaml_fl, list) else [nc_yaml_fl]:
            try:
                config = _from_yaml(yaml_fl)
            except RuntimeError as exc:
                raise RuntimeError from exc

            if "groups" in config:
                self.groups |= set(config["groups"])
            if "compounds" in config:
                self.compounds |= config["compounds"]
            if "dimensions" in config:
                self.dimensions |= config["dimensions"]
            if "variables" in config:
                self.variables |= config["variables"]
            if "attrs_global" in config:
                self.attrs_global |= config["attrs_global"]
            if "attrs_groups" in config:
                self.attrs_groups |= config["attrs_groups"]

    def __repr__(self: NcFromYaml) -> str:
        """Show object as dictionary."""
        return pprint.pformat(self.asdict)

    @property
    def asdict(self: NcFromYaml) -> dict:
        """Return definition of the HDF5/netCDF4 product."""
        return {
            "groups": self.groups,
            "dimensions": self.dimensions,
            "compounds": self.compounds,
            "variables": self.variables,
            "attrs_global": self.attrs_global,
            "attrs_groups": self.attrs_groups,
        }

    def use_netcdf4(self: NcFromYaml) -> NcFromYaml:
        """Use module netCDF4 to generate the HDF5/netCDF4 file."""
        if self.module == "netCDF4":
            return self

        self.module = "netCDF4"
        NcFromYaml.__bases__ = (NcCreate,)
        return self

    def use_h5py(self: NcFromYaml) -> NcFromYaml:
        """Use module h5py to generate the HDF5/netCDF4 file."""
        if self.module == "h5py":
            return self

        self.module = "h5py"
        NcFromYaml.__bases__ = (H5Create,)
        return self



#
# This file is part of Python package: `h5yaml`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025-2026 - R.M. van Hees (SRON)
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Initialize empty netCDF4 file using Python package `h5py`."""

from __future__ import annotations

__all__ = ["H5Create"]

import logging
from pathlib import Path

import h5py
import numpy as np

from . import __version__

H5_LIBVER = ("v110", "latest")


def str2bytes(ss: str) -> np.bytes_:
    """Convert Python string to ASCII encoded bytes."""
    return np.array(ss, dtype=h5py.string_dtype("ascii", len(ss)))


# - class definition -----------------------------------
class H5Create:
    """Class to create an empty netCDF4 file using Python package `h5py`.

    Parameters
    ----------
    groups: set | None = None
    compounds: dict | None = None
    dimensions: dict | None = None
    variables: dict | None = None
    attrs_global: dict | None = None
    attrs_groups: dict | None = None

    """

    str2bytes = False

    def __init__(
        self: H5Create,
        groups: set | None = None,
        compounds: dict | None = None,
        dimensions: dict | None = None,
        variables: dict | None = None,
        attrs_global: dict | None = None,
        attrs_groups: dict | None = None,
    ) -> None:
        """Construct a H5Create instance."""
        self.logger = logging.getLogger("h5yaml.H5Create")
        self.groups = set() if groups is None else groups
        self.compounds = {} if compounds is None else compounds
        self.dimensions = {} if dimensions is None else dimensions
        self.variables = {} if variables is None else variables
        self.attrs_global = {} if attrs_global is None else attrs_global
        self.attrs_groups = {} if attrs_groups is None else attrs_groups

    def _adjust_attr(
        self: H5Create, dtype: str, attr_key: str, attr_val: np.generic
    ) -> np.generic:
        """Return attribute converted to the same data type as its variable.

        Parameters
        ----------
        dtype :  str
           numpy data-type of variable
        attr_key :  str
           name of the attribute
        attr_val :  np.generic
           original value of the attribute

        Returns
        -------
        attr_val converted to dtype

        """
        if attr_key == "flag_values":
            return np.array(attr_val, dtype=dtype)

        if attr_key == "flag_masks":
            return np.array(attr_val, dtype=dtype)

        if attr_key == "scale_factor" and isinstance(attr_val, str):
            return eval(attr_val)

        if attr_key in ("valid_min", "valid_max", "valid_range"):
            match dtype:
                case "i1":
                    res = np.int8(attr_val)
                case "i2":
                    res = np.int16(attr_val)
                case "i4":
                    res = np.int32(attr_val)
                case "i8":
                    res = np.int64(attr_val)
                case "u1":
                    res = np.uint8(attr_val)
                case "u2":
                    res = np.uint16(attr_val)
                case "u4":
                    res = np.uint32(attr_val)
                case "u8":
                    res = np.uint64(attr_val)
                case "f2":
                    res = np.float16(attr_val)
                case "f4":
                    res = np.float32(attr_val)
                case "f8":
                    res = np.float64(attr_val)
                case _:
                    res = attr_val

            return res

        if self.str2bytes and isinstance(attr_val, str):
            return str2bytes(attr_val)

        return attr_val

    def __attrs(self: H5Create, fid: h5py.File) -> None:
        """Create global and group attributes.

        Parameters
        ----------
        fid :  h5py.File
           HDF5 file pointer (mode 'w' or 'r+')

        """
        value = (
            f"h5yaml.{self.__class__.__name__}(H5Create)"
            f",version={__version__.split('+', maxsplit=1)[0]}"
            f",{'options=str_as_bytes' if self.str2bytes else ''}"
        )
        fid.attrs["_NCCreator"] = str2bytes(value) if self.str2bytes else value
        value = f"version=2,hdf5={h5py.version.hdf5_version}"
        fid.attrs["_NCProperties"] = str2bytes(value) if self.str2bytes else value
        for key, value in self.attrs_global.items():
            if key in fid.attrs or value == "TBW":
                continue
            if isinstance(value, str):
                fid.attrs[key] = str2bytes(value) if self.str2bytes else value
            else:
                fid.attrs[key] = value

    def __groups(self: H5Create, fid: h5py.File) -> None:
        """Create groups in HDF5 product.

        Parameters
        ----------
        fid :  h5py.File
           HDF5 file pointer (mode 'w' or 'r+')

        """
        for key in self.groups:
            if key not in fid:
                _ = fid.create_group(key, track_order=True)
        for key, value in self.attrs_groups.items():
            if key in fid.attrs or value == "TBW":
                continue
            if isinstance(value, str):
                fid[str(Path(key).parent)].attrs[Path(key).name] = (
                    str2bytes(value) if self.str2bytes else value
                )
            else:
                fid[str(Path(key).parent)].attrs[Path(key).name] = value

    def __dimensions(self: H5Create, fid: h5py.File) -> None:
        """Add dimensions to HDF5 product.

        Parameters
        ----------
        fid :  h5py.File
         HDF5 file pointer (mode 'w' or 'r+')

        """
        for key, val in self.dimensions.items():
            fillvalue = None
            if "_FillValue" in val:
                fillvalue = np.nan if val["_FillValue"] == "NaN" else val["_FillValue"]

            if val["_size"] == 0:
                dset = fid.create_dataset(
                    key,
                    shape=(0,),
                    dtype="T" if val["_dtype"] == "str" else val["_dtype"],
                    chunks=True,
                    maxshape=(None,),
                    fillvalue=fillvalue,
                )
            else:
                dset = fid.create_dataset(
                    key,
                    shape=(val["_size"],),
                    dtype="T" if val["_dtype"] == "str" else val["_dtype"],
                    fillvalue=fillvalue,
                )
                if "_values" in val:
                    dset[:] = val["_values"]
                elif "_range" in val:
                    dset[:] = np.arange(*val["_range"], dtype=val["_dtype"])

            if fillvalue is not None:
                dset.attrs["_FillValue"] = dset.fillvalue

            dset.make_scale(
                Path(key).name
                if len(val) > 2
                else "This is a netCDF dimension but not a netCDF variable."
            )

            for attr, attr_val in val.items():
                if not attr.startswith("_"):
                    dset.attrs[attr] = self._adjust_attr(val["_dtype"], attr, attr_val)

    def __compounds(self: H5Create, fid: h5py.File) -> None:
        """Add compound datatypes to HDF5 product.

        Parameters
        ----------
        fid :  h5py.File
           HDF5 file pointer (mode 'w' or 'r+')

        """
        for key, val in self.compounds.items():
            fid[key] = np.dtype([(k, v[0]) for k, v in val.items()], align=True)

    def __var_scalar(self: H5Create, fid: h5py.File, key: str, val: dict) -> dict:
        """Return parameters to create a scalar variable.

        Parameters
        ----------
        key :  str
           Name of the variable
        val :  dict
           Properties of the variable

        """
        fillvalue = None
        if "_FillValue" in val:
            fillvalue = np.nan if val["_FillValue"] == "NaN" else val["_FillValue"]

        return {
            "name": key,
            "shape": (),
            "dtype": fid[val["_dtype"]].dtype
            if val["_dtype"] in fid
            else ("T" if val["_dtype"] == "str" else val["_dtype"]),
            "fillvalue": fillvalue,
        }

    def __var_nochunk(self: H5Create, fid: h5py.File, key: str, val: dict) -> dict:
        """Return parameters to create a variable without chunking.

        Parameters
        ----------
        key :  str
           Name of the variable
        val :  dict
           Properties of the variable

        """
        fillvalue = None
        if "_FillValue" in val:
            fillvalue = np.nan if val["_FillValue"] == "NaN" else val["_FillValue"]

        n_udim = 0
        ds_shape = ()
        for coord in val["_dims"]:
            dim_sz = fid[coord].size
            n_udim += int(dim_sz == 0)
            ds_shape += (dim_sz,)

        if n_udim > 0:
            raise KeyError(
                "you can not create a contiguous dataset with unlimited dimensions."
            )

        return {
            "name": key,
            "shape": ds_shape,
            "maxshape": None,
            "dtype": fid[val["_dtype"]].dtype
            if val["_dtype"] in fid
            else ("T" if val["_dtype"] == "str" else val["_dtype"]),
            "fillvalue": fillvalue,
        }

    def __var_chunked(self: H5Create, fid: h5py.File, key: str, val: dict) -> dict:
        """Return parameters to create a variable with chunking.

        Parameters
        ----------
        key :  str
           Name of the variable
        val :  dict
           Properties of the variable

        """
        fillvalue = None
        if "_FillValue" in val:
            fillvalue = np.nan if val["_FillValue"] == "NaN" else val["_FillValue"]

        n_udim = 0
        ds_shape = ()
        ds_maxshape = ()
        for coord in val["_dims"]:
            dim_sz = fid[coord].size
            n_udim += int(dim_sz == 0)
            ds_shape += (dim_sz,)
            ds_maxshape += (dim_sz if dim_sz > 0 else None,)

        # currently, we can not handle more than one unlimited dimension
        if n_udim > 1:
            raise ValueError(f"{key} has more than one unlimited dimension")

        ds_chunk = val.get("_chunks")
        if ds_chunk is not None and not isinstance(ds_chunk, bool):
            ds_chunk = tuple(ds_chunk)
        compression = None
        shuffle = False
        # currently only gzip compression is supported
        if "_compression" in val:
            compression = val["_compression"]
            shuffle = True

        ds_dtype = (
            fid[val["_dtype"]].dtype
            if val["_dtype"] in fid
            else ("T" if val["_dtype"] == "str" else val["_dtype"])
        )
        if "_vlen" in val:
            ds_name = (
                val["_dtype"].split("_")[0] if "_" in val["_dtype"] else val["_dtype"]
            ) + "_vlen"
            if ds_name not in fid:
                fid[ds_name] = h5py.vlen_dtype(ds_dtype)
                ds_dtype = fid[ds_name]
                fillvalue = None

        return {
            "name": key,
            "shape": ds_shape,
            "maxshape": ds_maxshape,
            "dtype": ds_dtype,
            "chunks": ds_chunk,
            "compression": compression,
            "shuffle": shuffle,
            "fillvalue": fillvalue,
        }

    def __variables(self: H5Create, fid: h5py.File) -> None:
        """Add datasets to HDF5 product.

        Parameters
        ----------
        fid :  h5py.File
           HDF5 file pointer (mode 'w' or 'r+')

        """
        for key, val in self.variables.items():
            # check if dtype of variable is compound
            is_compound = val["_dtype"] in fid

            # create variable
            is_scalar = False
            if val["_dims"][0] == "scalar":
                is_scalar = True
                dset = fid.create_dataset(**self.__var_scalar(fid, key, val))
            elif val.get("_chunks") == "contiguous":
                dset = fid.create_dataset(**self.__var_nochunk(fid, key, val))
            else:
                dset = fid.create_dataset(**self.__var_chunked(fid, key, val))

            if "_FillValue" in val and dset.fillvalue is not None:
                dset.attrs["_FillValue"] = dset.fillvalue

            for attr, attr_val in val.items():
                if attr.startswith("_"):
                    continue
                dset.attrs[attr] = self._adjust_attr(val["_dtype"], attr, attr_val)

            if is_compound:
                compound = self.compounds[val["_dtype"]]
                res = [v[2] for k, v in compound.items() if len(v) == 3]
                if res:
                    dset.attrs["units"] = [v[1] for k, v in compound.items()]
                    dset.attrs["names"] = res
                else:
                    dset.attrs["names"] = [v[1] for k, v in compound.items()]

            for ii, coord in enumerate([] if is_scalar else val["_dims"]):
                dset.dims[ii].attach_scale(fid[coord])

    def create(self: H5Create, filename: Path | str, str_as_bytes: bool = True) -> None:
        """Create a HDF5/netCDF4 file (overwrite if exist).

        Parameters
        ----------
        filename :  Path | str
           Full name of the HDF5/netCDF4 file to be generated
        str_as_bytes: bool, default=True
           Convert string to a netCDF4 compatable byte-array

        """
        self.str2bytes = str_as_bytes
        try:
            with h5py.File(filename, "w", track_order=True, libver=H5_LIBVER) as fid:
                self.__groups(fid)
                self.__dimensions(fid)
                self.__compounds(fid)
                self.__variables(fid)
                self.__attrs(fid)
        except PermissionError as exc:
            raise RuntimeError(f"failed create {filename}") from exc

    def diskless(self: H5Create, str_as_bytes: bool = True) -> h5py.File:
        """Create a HDF5/netCDF4 file in memory.

        Parameters
        ----------
        str_as_bytes: bool, default=True
           Convert string to a netCDF4 compatable byte-array

        Returns
        -------
          h5py.File: to add data to the empty HDF5 file

        """
        self.str2bytes = str_as_bytes
        fid = h5py.File.in_memory(track_order=True, libver=H5_LIBVER)
        self.__groups(fid)
        self.__dimensions(fid)
        self.__compounds(fid)
        self.__variables(fid)
        self.__attrs(fid)
        return fid

    def to_disk(self: H5Create, fid: h5py.File, filename: Path | str) -> None:
        """Write in-memory buffer to file.

        Parameters
        ----------
        fid :  h5py.File
           Dataset pointer returned by method `diskless`
        filename :  Path | str
           Name of the file on disk, or file-like object

        """
        fid.flush()
        try:
            with open(filename, "wb") as ff:
                _ = ff.write(fid.id.get_file_image())
        except PermissionError as exc:
            raise RuntimeError(f"failed create {filename}") from exc



#
# This file is part of Python package: `h5yaml`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025-2026 - R.M. van Hees (SRON)
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test module for h5yaml class `H5Yaml`."""

from __future__ import annotations

from importlib.resources import files
from pathlib import Path, PurePosixPath

import numpy as np
import pytest

from h5yaml.nc_from_yaml import NcFromYaml


class TestFromYaml:
    """Class to test H5Yaml from h5yaml.nc_from_yaml_h5."""

    _res = NcFromYaml(files("h5yaml.Data") / "nc_testing.yaml")
    NC_DEF = _res.asdict
    FID_NC = _res.use_netcdf4().diskless()

    def test_exceptions(self: TestFromYaml) -> None:
        """Unit-test for class exeptions."""
        yaml_path = files("h5yaml.Data") / "h5_testing.yaml"
        # raise an exception because folder dows not exist (str)
        l1a_name = "/this/folder/does/not/exists/test.nc"
        with pytest.raises(RuntimeError, match=r"failed to create .*") as excinfo:
            NcFromYaml(yaml_path).create(l1a_name)
        assert f"failed to create {l1a_name}" in str(excinfo)

        # raise an exception because the file can not be created (Path)
        l1a_name = Path("/this/folder/does/not/exists/test.nc")
        with pytest.raises(RuntimeError, match=r"failed to create .*") as excinfo:
            NcFromYaml(yaml_path).create(l1a_name)
        assert f"failed to create {l1a_name}" in str(excinfo)

        # raise exception due to permission error
        l1a_name = "/test.nc"
        with pytest.raises(RuntimeError, match=r"failed to create .*") as excinfo:
            NcFromYaml(yaml_path).create(l1a_name)
        assert f"failed to create {l1a_name}" in str(excinfo)

        # raise exception because YAML file can not be found
        yaml_path = Path("h5_testing.yaml")
        with pytest.raises(FileNotFoundError, match=r".* not found") as excinfo:
            _ = NcFromYaml(yaml_path).asdict
        assert f"{yaml_path} not found" in str(excinfo.value)

        # raise exception because the YAML file contains errors
        yaml_path = files("h5yaml.Data") / "h5_unsupported.yaml"
        with pytest.raises(ValueError, match=r".* unlimited dimension") as excinfo:
            _ = NcFromYaml(yaml_path).diskless()
        assert "more than one unlimited dimension" in str(excinfo)

    def test_nc_groups(self: TestFromYaml) -> None:
        """Unit-test to check the groups."""
        if "groups" not in self.NC_DEF:
            return

        for key in self.NC_DEF["groups"]:
            pkey = PurePosixPath(key)
            if len(pkey.parts) > 1:
                assert pkey.name in self.FID_NC[pkey.parent].groups
            else:
                assert key in self.FID_NC.groups

    def test_nc_dimensions(self: TestFromYaml) -> None:
        """Unit-test to check the dimensions."""
        if "dimensions" not in self.NC_DEF:
            return

        for key in self.NC_DEF["dimensions"]:
            pkey = PurePosixPath(key)
            dim_grp = self.FID_NC[pkey.parent] if pkey.is_absolute() else self.FID_NC
            dim_name = pkey.name if pkey.is_absolute() else key
            assert dim_name in dim_grp.dimensions

            if dim_name in dim_grp.variables:
                nc_dim = dim_grp.variables[dim_name]
                if "_dtype" in self.NC_DEF["dimensions"][key]:
                    dtype = self.NC_DEF["dimensions"][key]["_dtype"]
                    assert nc_dim.dtype == str if dtype == "str" else dtype
                if "_size" in self.NC_DEF["dimensions"][key]:
                    assert nc_dim.size == self.NC_DEF["dimensions"][key]["_size"]
                if "_values" in self.NC_DEF["dimensions"][key]:
                    assert np.array_equal(
                        nc_dim[:], self.NC_DEF["dimensions"][key]["_values"]
                    )
                if "_range" in self.NC_DEF["dimensions"][key]:
                    assert np.array_equal(
                        nc_dim[:], np.arange(*self.NC_DEF["dimensions"][key]["_range"])
                    )
                for attr in self.NC_DEF["dimensions"][key]:
                    if attr[0] == "_":
                        continue

                    assert getattr(nc_dim, attr) == self.NC_DEF["dimensions"][key][attr]

    def test_nc_compounds(self: TestFromYaml) -> None:
        """Unit-test to check the compounds."""
        if "compounds" not in self.NC_DEF:
            return

        for key in self.NC_DEF["compounds"]:
            assert key in self.FID_NC.cmptypes

    def test_nc_variables(self: TestFromYaml) -> None:
        """Unit-test to check the variables."""
        if "variables" not in self.NC_DEF:
            return

        for key in self.NC_DEF["variables"]:
            pkey = PurePosixPath(key)
            var_grp = self.FID_NC[pkey.parent] if pkey.is_absolute() else self.FID_NC
            var_name = pkey.name if pkey.is_absolute() else key
            assert var_name in var_grp.variables
            nc_var = var_grp.variables[var_name]

            if "_size" in self.NC_DEF["variables"][key]:
                assert nc_var.size == self.NC_DEF["variables"][key]["_size"]

            for attr in self.NC_DEF["variables"][key]:
                if attr[0] == "_":
                    continue

                if isinstance(
                    self.NC_DEF["variables"][key][attr], list | tuple | np.ndarray
                ):
                    assert np.array_equal(
                        getattr(nc_var, attr), self.NC_DEF["variables"][key][attr]
                    )
                else:
                    assert getattr(nc_var, attr) == self.NC_DEF["variables"][key][attr]

    def test_close(self: TestFromYaml) -> None:
        """Close the in-memory HDF5 file."""
        self.FID_NC.close()



#
# This file is part of Python package: `h5yaml`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025-2026 - R.M. van Hees (SRON)
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Initialize empty netCDF4 file using Python package `netCDF4`."""

from __future__ import annotations

__all__ = ["NcCreate"]

import logging
from pathlib import Path, PurePosixPath

import numpy as np

# pylint: disable=no-name-in-module
from netCDF4 import Dataset

from . import __version__
from .lib.adjust_attr import adjust_attr


# - class definition -----------------------------------
class NcCreate:
    """Class to create an empty netCDF4 file using Python package `netCDF4`.

    Parameters
    ----------
    groups: set | None = None
    compounds: dict | None = None
    dimensions: dict | None = None
    variables: dict | None = None
    attrs_global: dict | None = None
    attrs_groups: dict | None = None

    """

    def __init__(
        self: NcCreate,
        groups: set | None = None,
        compounds: dict | None = None,
        dimensions: dict | None = None,
        variables: dict | None = None,
        attrs_global: dict | None = None,
        attrs_groups: dict | None = None,
    ) -> None:
        """Construct a NcCreate instance."""
        self.logger = logging.getLogger("h5yaml.NcCreate")
        self.groups = set() if groups is None else groups
        self.compounds = {} if compounds is None else compounds
        self.dimensions = {} if dimensions is None else dimensions
        self.variables = {} if variables is None else variables
        self.attrs_global = {} if attrs_global is None else attrs_global
        self.attrs_groups = {} if attrs_groups is None else attrs_groups

    def __attrs(self: NcCreate, fid: Dataset) -> None:
        """Create global and group attributes.

        Parameters
        ----------
        fid :  netCDF4.Dataset
           netCDF4 Dataset (mode 'r+')

        """
        fid.setncattr(
            "_NCCreator",
            (
                f"h5yaml.{self.__class__.__name__}(NcCreate)"
                f",version={__version__.split('+', maxsplit=1)[0]}"
            ),
        )
        for key, val in self.attrs_global.items():
            if val == "TBW":
                continue

            fid.setncattr(key, val)

    def __groups(self: NcCreate, fid: Dataset) -> None:
        """Create groups in a netCDF4 product.

        Parameters
        ----------
        fid :  netCDF4.Dataset
           netCDF4 Dataset (mode 'r+')

        """
        for key in self.groups:
            _ = fid.createGroup(key)

        for key, val in self.attrs_groups.items():
            if val == "TBW":
                continue

            fid[str(Path(key).parent)].setncattr(Path(key).name, val)

    def __dimensions(self: NcCreate, fid: Dataset) -> None:
        """Add dimensions to a netCDF4 product.

        Parameters
        ----------
        fid :  netCDF4.Dataset
           netCDF4 Dataset (mode 'r+')

        """
        for key, value in self.dimensions.items():
            pkey = PurePosixPath(key)
            if pkey.is_absolute():
                _ = fid[pkey.parent].createDimension(pkey.name, value["_size"])
            else:
                _ = fid.createDimension(key, value["_size"])

            if len(value) <= 2:
                continue

            fillvalue = None
            if "_FillValue" in value:
                fillvalue = (
                    np.nan if value["_FillValue"] == "NaN" else value["_FillValue"]
                )

            if pkey.is_absolute():
                dset = fid[pkey.parent].createVariable(
                    pkey.name,
                    value["_dtype"],
                    dimensions=(pkey.name,),
                    fill_value=fillvalue,
                    contiguous=value["_size"] != 0,
                )
            else:
                dset = fid.createVariable(
                    key,
                    value["_dtype"],
                    dimensions=(key,),
                    fill_value=fillvalue,
                    contiguous=value["_size"] != 0,
                )
            if value["_size"] > 0:
                if "_values" in value:
                    dset[:] = np.array(value["_values"])
                elif "_range" in value:
                    dset[:] = np.arange(*value["_range"], dtype=value["_dtype"])

            dset.setncatts(
                {
                    k: adjust_attr(value["_dtype"], k, v)
                    for k, v in value.items()
                    if not k.startswith("_")
                }
            )

    def __compounds(self: NcCreate, fid: Dataset) -> None:
        """Add compound datatypes to a netCDF4 product.

        Parameters
        ----------
        fid :  netCDF4.Dataset
           netCDF4 Dataset (mode 'r+')

        """
        for key, val in self.compounds.items():
            comp_t = np.dtype([(k, v[0]) for k, v in val.items()])
            _ = fid.createCompoundType(comp_t, key)

    def __var_scalar(self: NcCreate, fid: Dataset, key: str, val: dict) -> dict:
        """Return parameters to create a scalar variable.

        Parameters
        ----------
        key :  str
           Name of the variable
        val :  dict
           Properties of the variable

        """
        fillvalue = None
        if "_FillValue" in val:
            fillvalue = np.nan if val["_FillValue"] == "NaN" else int(val["_FillValue"])

        return {
            "varname": key,
            "datatype": (
                fid.cmptypes[val["_dtype"]]
                if val["_dtype"] in fid.cmptypes
                else val["_dtype"]
            ),
            "contiguous": True,
            "fill_value": fillvalue,
        }

    def __var_nochunk(self: NcCreate, fid: Dataset, key: str, val: dict) -> dict:
        """Return parameters to create a variable without chunking.

        Parameters
        ----------
        key :  str
           Name of the variable
        val :  dict
           Properties of the variable

        """
        fillvalue = None
        if "_FillValue" in val:
            fillvalue = np.nan if val["_FillValue"] == "NaN" else int(val["_FillValue"])

        n_udim = 0
        var_dims = []
        for coord in val["_dims"]:
            pcoord = PurePosixPath(coord)
            var_dims.append(pcoord.name if pcoord.is_absolute() else coord)
            if pcoord.is_absolute():
                dim_sz = fid[pcoord.parent].dimensions[pcoord.name].size
            else:
                dim_sz = fid.dimensions[coord].size
            n_udim += int(dim_sz == 0)

        # currently, we can not handle more than one unlimited dimension
        if n_udim > 0:
            raise KeyError(
                "you can not create a contiguous dataset with unlimited dimensions."
            )

        return {
            "varname": key,
            "datatype": (
                fid.cmptypes[val["_dtype"]]
                if val["_dtype"] in fid.cmptypes
                else val["_dtype"]
            ),
            "dimensions": var_dims,
            "contiguous": True,
            "fill_value": fillvalue,
        }

    def __var_chunked(self: NcCreate, fid: Dataset, key: str, val: dict) -> dict:
        """Return parameters to create a variable with chunking.

        Parameters
        ----------
        key :  str
           Name of the variable
        val :  dict
           Properties of the variable

        """
        fillvalue = None
        if "_FillValue" in val:
            fillvalue = np.nan if val["_FillValue"] == "NaN" else int(val["_FillValue"])

        n_udim = 0
        var_dims = []
        for coord in val["_dims"]:
            pcoord = PurePosixPath(coord)
            var_dims.append(pcoord.name if pcoord.is_absolute() else coord)
            if pcoord.is_absolute():
                dim_sz = fid[pcoord.parent].dimensions[pcoord.name].size
            else:
                dim_sz = fid.dimensions[coord].size
            n_udim += int(dim_sz == 0)

        # currently, we can not handle more than one unlimited dimension
        if n_udim > 1:
            raise ValueError(f"{key} has more than one unlimited dimension")

        compression = None
        complevel = 0
        # currently only gzip compression is supported
        if "_compression" in val:
            compression = "zlib"
            complevel = val["_compression"]

        ds_chunk = val.get("_chunks")
        if ds_chunk is not None:
            ds_chunk = None if isinstance(ds_chunk, bool) else tuple(ds_chunk)

        ds_dtype = (
            fid.cmptypes[val["_dtype"]]
            if val["_dtype"] in fid.cmptypes
            else val["_dtype"]
        )
        if "_vlen" in val:
            if val["_dtype"] in fid.cmptypes:
                raise ValueError("can not have vlen with compounds")
            ds_dtype = fid.createVLType(ds_dtype, "phony_vlen")
            fillvalue = None

        return {
            "varname": key,
            "datatype": ds_dtype,
            "dimensions": var_dims,
            "fill_value": fillvalue,
            "compression": compression,
            "complevel": complevel,
            "chunksizes": ds_chunk,
        }

    def __variables(self: NcCreate, fid: Dataset) -> None:
        """Add datasets to a netCDF4 product.

        Parameters
        ----------
        fid :  netCDF4.Dataset
           netCDF4 Dataset (mode 'r+')

        """
        for key, val in self.variables.items():
            pkey = PurePosixPath(key)
            var_grp = fid[pkey.parent] if pkey.is_absolute() else fid
            var_name = pkey.name if pkey.is_absolute() else key

            # check if dtype of variable is compound
            is_compound = val["_dtype"] in fid.cmptypes

            # create variable
            if val["_dims"][0] == "scalar":
                dset = var_grp.createVariable(**self.__var_scalar(fid, var_name, val))
            elif val.get("_chunks") == "contiguous":
                dset = var_grp.createVariable(**self.__var_nochunk(fid, var_name, val))
            else:
                dset = var_grp.createVariable(**self.__var_chunked(fid, var_name, val))

            dset.setncatts(
                {
                    k: adjust_attr(val["_dtype"], k, v)
                    for k, v in val.items()
                    if not k.startswith("_")
                }
            )
            if is_compound:
                compound = self.compounds[val["_dtype"]]
                res = [v[2] for k, v in compound.items() if len(v) == 3]
                if res:
                    dset.units = [v[1] for k, v in compound.items()]
                    dset.names = res
                else:
                    dset.names = [v[1] for k, v in compound.items()]

    def create(self: NcCreate, filename: Path | str) -> None:
        """Create a netCDF4 file (overwrite if exist).

        Parameters
        ----------
        filename :  Path | str
           Name of the file on disk, or file-like object

        """
        try:
            with Dataset(filename, "w") as fid:
                self.__groups(fid)
                self.__compounds(fid)
                self.__dimensions(fid)
                self.__variables(fid)
                self.__attrs(fid)
        except PermissionError as exc:
            raise RuntimeError(f"failed to create {filename}") from exc

    def diskless(self: NcCreate) -> Dataset:
        """Create a netCDF4 file in memory.

        Returns
        -------
          Dataset: to add data to the empty netCDF4 file.

        """
        fid = Dataset("diskless.nc", "w", memory=2**30)
        self.__groups(fid)
        self.__compounds(fid)
        self.__dimensions(fid)
        self.__variables(fid)
        self.__attrs(fid)
        return fid

    def to_disk(self: NcCreate, fid: Dataset, filename: Path | str) -> None:
        """Write in-memory buffer to file, and close the Dataset.

        Parameters
        ----------
        fid :  Dataset
           Dataset pointer returned by method `diskless`
        filename :  Path | str
           Name of the file on disk, or file-like object

        """
        try:
            with open(filename, "wb") as ff:
                _ = ff.write(fid.close())
        except PermissionError as exc:
            raise RuntimeError(f"failed create {filename}") from exc
        except OSError as exc:
            raise RuntimeError(f"failed to write {filename}") from exc
